import numpy as np
from lazylinop import binary_dtype, LazyLinOp
from lazylinop.basicops import zeros


def anti_eye(M: int, N: int = None, k: int = 0, dtype: str = 'float'):
    """Constructs a :class:`.LazyLinOp` whose equivalent array is filled with
    ones on the k-th antidiagonal and zeros everywhere else.

    ``L = anti_eye(M, N, k)`` is such that ``L.toarray() ==
    numpy.flip(numpy.eye(M, N, k), axis=1)``.

    Args:
        M: ``int``
            Number of rows.
        N: ``int``, optional
            Number of columns (default is ``M``).
        k: ``int``, optional
            Anti-diagonal to place ones on.

            - zero for the main antidiagonal (default),
              the one starting from the upper right corner.
            - positive integer for an upper antidiagonal.
            - negative integer for a lower antidiagonal,

            if ``k >= N`` or ``k <= - M`` then ``anti_eye(M, N, k)`` is
            ``zeros((M, N))`` (k-th antidiagonal is out of operator shape).

        dtype: ``str`` or ``numpy.dtype``, optional
            Defaultly ``'float'``.

    Returns:
        The anti-eye :class:`.LazyLinOp`.

    Examples:
        >>> import lazylinop as lz
        >>> import numpy as np
        >>> x = np.arange(3)
        >>> L = lz.basicops.anti_eye(3)
        >>> np.allclose(L @ x, np.flip(x))
        True
        >>> # Check the main diagonal is the one starting
        >>> # from the upper right corner.
        >>> L = lz.basicops.anti_eye(3, 4, k=0)
        >>> L.toarray()
        array([[0., 0., 0., 1.],
               [0., 0., 1., 0.],
               [0., 1., 0., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=3, k=0)
        >>> L.toarray()
        array([[0., 0., 1.],
               [0., 1., 0.],
               [1., 0., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=3, k=1)
        >>> L.toarray()
        array([[0., 1., 0.],
               [1., 0., 0.],
               [0., 0., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=3, k=-1)
        >>> L.toarray()
        array([[0., 0., 0.],
               [0., 0., 1.],
               [0., 1., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=4, k=0)
        >>> L.toarray()
        array([[0., 0., 0., 1.],
               [0., 0., 1., 0.],
               [0., 1., 0., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=4, k=1)
        >>> L.toarray()
        array([[0., 0., 1., 0.],
               [0., 1., 0., 0.],
               [1., 0., 0., 0.]])
        >>> L = lz.basicops.anti_eye(3, N=4, k=-1)
        >>> L.toarray()
        array([[0., 0., 0., 0.],
               [0., 0., 0., 1.],
               [0., 0., 1., 0.]])

    .. seealso::
        :py:func:`.eye`,
        :py:func:`.LazyLinOp.toarray`,
        `numpy.eye
        <https://numpy.org/doc/stable/reference/generated/numpy.eye.html>`_,
        `numpy.flip
        <https://numpy.org/doc/stable/reference/generated/numpy.flip.html>`_.
    """
    nn = M if N is None else N
    if M < 1 or nn < 1:
        raise ValueError("M and N must be >= 1.")

    if k >= nn or k <= - M:
        # diagonal is out of shape
        # return zeros LazyLinOp
        return zeros((M, nn))

    def _matmat(x, M, N, k):
        # x is always 2d
        out_dtype = binary_dtype(dtype, x.dtype)
        y = np.zeros((M, x.shape[1]), dtype=x.dtype)
        if k == 0:
            # Main anti-diagonal
            nc = min(M, N)
            y[:nc, :] = x[-1:- 1 - nc:-1]
        elif k > 0:
            # Above anti-diagonal
            # k <= N
            nc = max(0, min(M, N - k))
            y[:nc] = x[-1 - k:-1 - k - nc:-1]
        else:
            # Below anti-diagonal (k < 0)
            nc = max(0, min(M + k, N))
            y[-k:-k+nc] = x[-1:- 1 - nc:-1]
        return y.astype(out_dtype)

    def _rmatmat(x, N, M, k):
        out_dtype = binary_dtype(dtype, x.dtype)
        y = np.zeros((N, x.shape[1]), dtype=x.dtype)
        if k == 0:
            # Main anti-diagonal transpose
            y[max(N - M, 0):] = x[min(M - 1, N - 1)::-1]
        elif k > 0:
            # Above anti-diagonal transpose
            if M >= N:
                # k < N
                nc = N - k
                y[:nc] = x[nc - 1::-1]
            else:
                # M < N
                ys = max(0, N - M - k)
                nc = min(N - k - 1, M - 1)
                y[ys:ys + nc + 1] = x[nc::-1]
        else:
            # Below anti-diagonal (k < 0) transpose
            mpk = M + k
            if N == M:
                nc = max(0, min(mpk, N))
                y[-k:-k+nc] = x[-1:- 1 - nc:-1]
            elif M > N:
                nc = min(N, mpk)  # number entries to copy
                xs = M-1-max(mpk-N, 0)  # x read start
                ys = max(N - nc, 0)  # y write start
                y[ys:nc + ys] = x[xs:xs - nc:-1]
            else:
                y[N-M-k:] = x[M-1:-k-1:-1]

        return y.astype(out_dtype)

    return LazyLinOp(
        shape=(M, nn),
        matmat=lambda x: _matmat(x, M, nn, k),
        rmatmat=lambda x: _rmatmat(x, nn, M, k),
        dtype=dtype
    )
